from __future__ import annotations
from bagua.torch_api.communication import get_backend
import bagua
from bagua.torch_api.utils import to_bagua_datatype, StatisticalAverage
from bagua.torch_api.env import get_autotune_level, get_rank
from bagua.bagua_define import (
    TensorDeclaration,
    BaguaHyperparameter,
)
import gorilla
import time
import logging
import torch
import torch.nn
import itertools
from typing import List, Tuple


@gorilla.patches(torch.nn.Module, filter=lambda name, obj: "bagua" in name)
class BaguaModule:
    """
    This class patches `torch.nn.Module` with several methods to enable Bagua
    functionalities.

    :ivar bagua_optimizers: The optimizers passed in by ``with_bagua(...)``.
    :vartype bagua_optimizers: List[torch.optim.Optimizer]

    :ivar bagua_algorithm: The algorithm passed in by ``with_bagua(...)``.
    :vartype bagua_algorithm: bagua.torch_api.algorithms.Algorithm

    :ivar parameters_to_ignore: The parameter names in ``"{module_name}.{param_name}"`` format to ignore
        when calling ``self.bagua_build_params()``.
    :vartype parameters_to_ignore: List[str]

    :ivar bagua_train_step_counter: Number of iterations in training mode
    :vartype bagua_train_step_counter: int

    :ivar bagua_buckets: All Bagua buckets in a list.
    :vartype bagua_buckets: List[bagua.torch_api.bucket.BaguaBucket]
    """

    __id_iter = itertools.count()

    def bagua_build_params(self) -> List[Tuple[str, torch.nn.Parameter]]:
        """
        Build tuple of ``(parameter_name, parameter)`` for all parameters that
        require grads and not in the ``_bagua_params_and_buffers_to_ignore`` attribute.
        """
        modules_and_parameters = [
            (module, parameter)
            for module_name, module in self.named_modules()
            for parameter in [
                (f"{module_name}.{param_name}", param)
                # Note that we access module.named_parameters instead of
                # parameters(module). parameters(module) is only needed in the
                # single-process multi device case, where it accesses replicated
                # parameters through _former_parameters.
                for param_name, param in module.named_parameters(recurse=False)
                if param.requires_grad
                and f"{module_name}.{param_name}" not in self.parameters_to_ignore
            ]
        ]

        # Deduplicate any parameters that might be shared across child modules.
        memo = set()
        # "p not in memo" is the deduplication check.
        # "not memo.add(p)" is always True, and it's only there to cause "add(p)" if needed.
        modules_and_parameters = [
            (m, p)
            for m, p in modules_and_parameters
            if p not in memo and not memo.add(p)
        ]

        # Build list of parameters.
        parameters = [parameter for _, parameter in modules_and_parameters]

        # Checks if a module will produce a sparse gradient.
        def produces_sparse_gradient(module):
            if isinstance(module, torch.nn.Embedding) or isinstance(
                module, torch.nn.EmbeddingBag
            ):
                return module.sparse
            return False

        # Build list of booleans indicating whether or not to expect sparse
        # gradients for the corresponding parameters.
        expect_sparse_gradient = [
            produces_sparse_gradient(module) for module, _ in modules_and_parameters
        ]

        if any(expect_sparse_gradient):
            raise NotImplementedError("sparse gradient not supported yet")

        return parameters

    def _bagua_broadcast_parameters(self):
        """
        Broadcast model and optimizer states.
        """
        from bagua.torch_api.communication import broadcast

        module_states = self.bagua_build_params()
        for name, state in module_states:
            broadcast(state, root=0)
        for optimizer in self.bagua_optimizers:
            optimizer_state_dict = optimizer.state_dict()["state"]
            for state in optimizer_state_dict.values():
                for inner_state in state.values():
                    if isinstance(
                        inner_state, torch.Tensor
                    ):  # TODO: consider the case where this is a scalar
                        broadcast(inner_state, root=0)

    def _bagua_autotune_step(self):
        CYCLE_STEP = 100
        start_time = time.time()

        if (
            self.bagua_train_step_counter != 0
            and self.bagua_train_step_counter % CYCLE_STEP == 0
        ):
            # get speed metrics
            time_since_last_update = time.time() - self._bagua_autotune_last_report_time
            speed = self._speed_metrics.get(time_since_last_update)

            # report metrics
            # TODO: @shjwudp add support for reporting tensor completion order
            # so that the autotune service does not rely on tensor registration
            # order
            rsp = self._bagua_autotune_client.report_metrics(
                model_name=self.bagua_module_name,
                rank=get_rank(),
                unix_timestamp=time.time(),
                train_iter=self.bagua_train_step_counter,
                hyperparameters=self._bagua_hyperparameters.dict(),
                speed=speed,
            )
            assert rsp.status_code == 200, "Unexpected rsp={}".format(rsp)

            # update parameters
            self._bagua_reset_algorithm_buckets()
            self._bagua_autotune_last_report_time = time.time()

        logging.info("autotune overhead=%s", time.time() - start_time)

    def with_bagua(  # pytype: disable=module-attr
        self,
        optimizers: List[torch.optim.Optimizer],
        algorithm: "bagua.torch_api.algorithms.Algorithm",
    ) -> BaguaModule:
        r"""``with_bagua`` enables easy distributed data parallel training on a
        ``torch.nn.Module``.

        Arguments:
            optimizers: Optimizer(s) used by the
                module. It can contain one or more PyTorch optimizers.
            algorithm: Distributed algorithm
                used to do the actual communication and update.

        Returns:
            The original module, with Bagua related environments initialized.

        .. note::
            If we want to ignore some layers for communication, we can first check
            these layer's corresponding keys in the module's ``state_dict`` (they are
            in ``"{module_name}.{param_name}"`` format), then assign the list of
            keys to ``your_module._bagua_params_and_buffers_to_ignore``.

        Examples::

            >>> model = torch.nn.Sequential(
            ...      torch.nn.Linear(D_in, H),
            ...      torch.nn.ReLU(),
            ...      torch.nn.Linear(H, D_out),
            ...    )
            >>> optimizer = torch.optim.SGD(
            ...      model.parameters(),
            ...      lr=0.01,
            ...      momentum=0.9
            ...    )
            >>> model = model.with_bagua(
            ...      [optimizer],
            ...      GradientAllReduce()
            ...    )
        """

        self.bagua_module_name = "{}_{}".format(
            self.__class__.__name__, next(BaguaModule.__id_iter)
        )

        self.bagua_optimizers = optimizers
        self.bagua_algorithm = algorithm
        self.parameters_to_ignore = (
            []
        )  #: the parameter names to ignore during communication
        if hasattr(self, "_bagua_params_and_buffers_to_ignore"):
            self.parameters_to_ignore.extend(self._bagua_params_and_buffers_to_ignore)
        if hasattr(
            self, "_ddp_params_and_buffers_to_ignore"
        ):  # for compatibility with PyTorch DDP
            self.parameters_to_ignore.extend(self._ddp_params_and_buffers_to_ignore)
        self.bagua_train_step_counter = 0
        """
        Number of iterations in training mode.
        """
        self.bagua_buckets = []
        """
        All Bagua buckets in a list.
        """
        self._bagua_autotune_last_report_time = time.time()
        self._bagua_autotune_completed = False
        self._bagua_framework_hooks = (
            []
        )  # hooks for bagua framework logic, not cleared when changing algorithms
        self._bagua_algorithm_hooks = []
        self._bagua_backend = get_backend(self.bagua_module_name)
        self._bagua_hyperparameters = BaguaHyperparameter()
        self._speed_metrics_switch_on = get_autotune_level() >= 1
        self._speed_metrics = StatisticalAverage()

        def autotune_hook(self, input):
            if self.training:
                if get_autotune_level() >= 1 and not self._bagua_autotune_completed:
                    self._bagua_autotune_step()

        def clear_post_backward_callback_queued_hook(self, input):
            self._is_post_backward_callback_queued = False

        def num_iteration_step_hook(self, input):
            if self.training:
                self.bagua_train_step_counter += 1

        def algorithm_reset_hook(self, input):
            if self.bagua_algorithm.need_reset():
                self._bagua_init_algorithm()

        def algorithm_forward_pre_hook(self, input):
            self.bagua_algorithm.init_forward_pre_hook(self)(input)

        def record_speed_metrics_event(self, _):
            if not self._speed_metrics_switch_on:
                return

            if hasattr(self, "_last_event_pair"):
                (start, stop) = self._last_event_pair
                try:
                    elapsed_time_s = start.elapsed_time(stop) / 1000.0
                    total_bytes = sum(bucket.bytes() for bucket in self.bagua_buckets)
                    total_gbytes = total_bytes / 1024.0 ** 3
                    speed = total_gbytes / elapsed_time_s
                    self._speed_metrics.record(speed)
                except RuntimeError as err:
                    logging.debug("Ignore cuda err={}".format(err))

            start_event = torch.cuda.Event(enable_timing=True)
            self._speed_metrics_end_event = torch.cuda.Event(enable_timing=True)
            torch.cuda.current_stream().record_event(start_event)
            self._last_event_pair = (start_event, self._speed_metrics_end_event)

        self._bagua_framework_hooks.extend(
            [
                self.register_forward_pre_hook(num_iteration_step_hook),
                self.register_forward_pre_hook(algorithm_reset_hook),
                self.register_forward_pre_hook(algorithm_forward_pre_hook),
                self.register_forward_pre_hook(record_speed_metrics_event),
                self.register_forward_pre_hook(autotune_hook),
                self.register_forward_pre_hook(
                    clear_post_backward_callback_queued_hook
                ),
            ]
        )

        # get communicators
        self._bagua_inter_node_communicator = (
            self._bagua_backend.internode_communicator
        )
        self._bagua_intra_node_communicator = (
            self._bagua_backend.intranode_communicator
        )
        self._bagua_global_communicator = self._bagua_backend.global_communicator
        self.bagua_communication_stream = self._bagua_backend.stream

        # autotune service
        from bagua.torch_api.communication import get_hyperparameters_service_client

        self._bagua_autotune_client = get_hyperparameters_service_client()

        self._bagua_init_algorithm()
        return self

    def _bagua_autotune_register_tensors(self):
        """
        Register tensors on autotune server, and return first bucketing suggestions
        """
        autotune_tensor_list = [
            TensorDeclaration(
                {
                    "name": tensor.bagua_tensor_name,
                    "num_elements": tensor.numel(),
                    "dtype": to_bagua_datatype(tensor.dtype),
                }
            )
            for tensor in self._bagua_tensors
        ]

        rsp = self._bagua_autotune_client.register_tensors(
            model_name=self.bagua_module_name, tensor_list=autotune_tensor_list
        )
        assert rsp.status_code == 200, "Unexpected rsp={}".format(rsp)

    def _bagua_autotune_get_buckets(self):
        rsp = self._bagua_autotune_client.ask_hyperparameters(
            model_name=self.bagua_module_name,
            rank=get_rank(),
            train_iter=self.bagua_train_step_counter,
        )
        assert rsp.status_code == 200, "Unexpected rsp={}".format(rsp)
        recommended_hyperparameters = rsp.json()["recommended_hyperparameters"]
        is_autotune_completed = rsp.json()["is_autotune_completed"]

        self._bagua_hyperparameters.update(recommended_hyperparameters)

        self._bagua_autotune_completed = is_autotune_completed
        recommended_buckets = map(
            lambda x: list(map(lambda y: self._bagua_tensor_map[y["name"]], x)),
            recommended_hyperparameters["buckets"],
        )
        return list(recommended_buckets)

    def _bagua_init_algorithm(self):
        self._bagua_cleanup_algorithm()
        self._bagua_broadcast_parameters()
        self._bagua_tensors = self.bagua_algorithm.init_tensors(self)
        self._bagua_tensor_map = dict(
            [(tensor.bagua_tensor_name, tensor) for tensor in self._bagua_tensors]
        )
        self._bagua_autotune_register_tensors()
        self._bagua_reset_algorithm_buckets()

    def _bagua_cleanup_algorithm(self):
        for hook in self._bagua_algorithm_hooks:
            hook.remove()
        self._bagua_algorithm_hooks.clear()
        self.bagua_buckets.clear()

    def _bagua_reset_algorithm_buckets(self):
        self._bagua_cleanup_algorithm()
        raw_buckets = self._bagua_autotune_get_buckets()
        self.bagua_buckets.extend(self.bagua_algorithm.tensors_to_buckets(raw_buckets))

        for name, param in self.named_parameters():

            def real_hook_factory(param_name, parameter):
                def real_hook(*unused):
                    self.bagua_algorithm.init_backward_hook(self)(param_name, parameter)

                    def real_post_backward_hook(*unused):
                        if self._speed_metrics_switch_on:
                            torch.cuda.current_stream().record_event(
                                self._speed_metrics_end_event
                            )

                        self.bagua_algorithm.init_post_backward_hook(self)()

                    if not self._is_post_backward_callback_queued:
                        torch.autograd.Variable._execution_engine.queue_callback(
                            real_post_backward_hook
                        )
                        self._is_post_backward_callback_queued = True

                return real_hook

            if param.requires_grad:
                param_tmp = param.expand_as(param)
                grad_acc = param_tmp.grad_fn.next_functions[0][0]
                hook = grad_acc.register_hook(real_hook_factory(name, param))
                hook.grad_acc = grad_acc
                self._bagua_algorithm_hooks.append(hook)

        optimizer_hook = self.bagua_algorithm.init_post_optimizer_step_hook(self)

        from types import MethodType

        for optimizer in self.bagua_optimizers:
            if not hasattr(optimizer, "_bagua_original_step"):
                optimizer._bagua_original_step = optimizer.step

            def new_step_factory(optimizer):
                def new_step(self, *args, **kwargs):
                    result = self._bagua_original_step(*args, **kwargs)
                    optimizer_hook(self)
                    return result

                return MethodType(new_step, optimizer)

            optimizer.step = new_step_factory(optimizer)

        for bucket in self.bagua_buckets:
            self.bagua_algorithm.init_operations(
                self,
                bucket,
            )
        self._bagua_backend.register_ordered_buckets(
            [bucket.backend_bucket for bucket in self.bagua_buckets]
        )


_base = gorilla._get_base(BaguaModule)
_decorator_data = gorilla.get_decorator_data(_base)
for patch in _decorator_data.patches:
    gorilla.apply(patch)



# TODO: @shjwudp merge with service module
import copy
import collections
import logging
import os
import re
import skopt
import time
from typing import Tuple, Optional
from pssh.clients import ParallelSSHClient
from pssh.exceptions import Timeout


class IntParam:
    def __init__(self, val: int, space_dimension: Tuple[int, int]):
        self.val: int = val
        self.space_dimension: Tuple[int, int] = space_dimension

    def __str__(self):
        return str(self.__dict__)


class FloatParam:
    def __init__(self, val: float, space_dimension: Tuple[float, float]):
        self.val: float = val
        self.space_dimension: Tuple[float, float] = space_dimension

    def __str__(self):
        return str(self.__dict__)


class BoolParam:
    def __init__(self, val: bool):
        self.val: bool = val
        self.space_dimension: Tuple[int, int] = (0, 1)

    def __str__(self):
        return str(self.__dict__)


class BayesianOptimizer:
    """
    Simple package of beyasian optimizer
    """

    def __init__(
        self,
        param_declaration: dict,
        n_initial_points: int = 20,
        initial_point_generator: str = "halton",
        random_state: Optional[int] = 0,
    ):
        self.param_declaration = collections.OrderedDict(param_declaration)
        search_space = [
            declar.space_dimension for _, declar in self.param_declaration.items()
        ]

        self.bayesian_optimizer = skopt.Optimizer(
            dimensions=search_space,
            n_initial_points=n_initial_points,
            initial_point_generator=initial_point_generator,
            n_jobs=-1,
            random_state=random_state,
        )

    def tell(self, param_dict: dict, score: float) -> None:
        param_v = [
            float(param_dict[name])
            for name, _ in self.param_declaration.items()  # noqa: E501
        ]
        try:
            self.bayesian_optimizer.tell(param_v, -score)
        except ValueError as err:
            logging.warning(
                "Maybe sklearn's division by 0 bug, skip it. err={}, param_v={}".format(
                    err, param_v
                )
            )

    def ask(self) -> dict:
        param_v = self.bayesian_optimizer.ask()
        param_dict = {}
        for i, (name, _) in enumerate(self.param_declaration.items()):
            param_dict[name] = param_v[i]

        return param_dict


def sysperf(host_list, nproc_per_node, ssh_port, env: dict = {}):
    assert len(host_list) != 0, "Invalid host_list={}".format(host_list)

    if "PATH" not in env:
        env["PATH"] = os.environ["PATH"]

    pretreat_cmd = [
        "shopt -s huponexit &&",
    ]
    for k, v in env.items():
        pretreat_cmd.append(
            "{key}={value} &&".format(
                key=k,
                value=v,
            )
        )

    master_addr = host_list[0]
    host_args = []
    for i, _ in enumerate(host_list):
        host_args.append(
            {
                "cmd": " ".join(
                    pretreat_cmd
                    + [
                        "python -m bagua.distributed.launch",
                        "--nproc_per_node={}".format(nproc_per_node),
                        "--nnodes={} --node_rank={}".format(len(host_list), i),
                        '--master_addr="{}"'.format(master_addr),
                        "--master_port={}".format(8124),
                        "$(which bagua_sys_perf) --model vgg16",
                    ]
                ),
            }
        )

    client = ParallelSSHClient(host_list, port=ssh_port)
    output = client.run_command(
        "%(cmd)s",
        host_args=host_args,
        shell="bash -xc",
        use_pty=True,  # The key configuration of process safe exit
        read_timeout=60,
    )
    speed_pattern = re.compile(
        r"Total img/sec on (\d+) (\S+)\(s\): (\d*\.\d+|\d+) \+-(\d*\.\d+|\d+)"
    )
    host_out = output[0]
    m = None

    st = time.time()
    try:
        for line in host_out.stdout:
            print(line, flush=True)
            m = speed_pattern.search(line)
            if m:
                break
    except Timeout:
        print("Timeout 1, spend={}".format(time.time() - st))
        pass

    if m is None:
        return (None, None, 0.0, None)

    assert m, "no speed pattern, host_out.exit_code={}, host_out.stderr={}".format(
        host_out.exit_code, list(host_out.stderr)
    )

    ngpus = int(m.groups()[0])
    device = m.groups()[1]
    speed = float(m.groups()[2])
    speed_std = float(m.groups()[3])

    return (ngpus, device, speed, speed_std)


def autotune_system_hyperparameters(host_list, nproc_per_node, ssh_port):
    def _sysperf(env={}):
        result = sysperf(host_list, nproc_per_node, ssh_port, env=env)
        print(result)
        return result

    optim = BayesianOptimizer(
        {
            "NCCL_MIN_NCHANNELS": IntParam(
                val=0,  # 0 means no set
                space_dimension=(
                    0,
                    12,
                ),
            ),
            "NCCL_SOCKET_NTHREADS": IntParam(
                val=0,  # 0 means no set
                space_dimension=(
                    0,
                    8,
                ),
            ),
            "NCCL_NSOCKS_PERTHREAD": IntParam(
                val=0,  # 0 means no set
                space_dimension=(
                    0,
                    8,
                ),
            ),
            "nccl_buffsize_2p": IntParam(
                val=0,  # power of 2, 0 means no set
                space_dimension=(
                    0,
                    26,
                ),
            ),
        }
    )

    param_dict = {
        "NCCL_MIN_NCHANNELS": 0,
        "NCCL_SOCKET_NTHREADS": 0,
        "NCCL_NSOCKS_PERTHREAD": 0,
        "nccl_buffsize_2p": 0,
    }

    result_list = []
    for i in range(100):
        env_vars = copy.deepcopy(param_dict)
        for k, v in list(env_vars.items()):
            if v == 0:
                del env_vars[k]

        if "nccl_buffsize_2p" in env_vars:
            env_vars["NCCL_BUFFSIZE"] = 2 ** env_vars["nccl_buffsize_2p"]

        (_, _, speed, speed_std) = _sysperf(env=env_vars)
        result_list.append([copy.deepcopy(env_vars), speed, speed_std])
        optim.tell(param_dict, speed)
        param_dict = optim.ask()

    result_list = sorted(result_list, key=lambda x: -x[1])
    print(result_list)

    result_reduct = {}
    for (setting, speed, _) in result_list:
        key = tuple(sorted(setting.items()))
        if key not in result_reduct:
            result_reduct[key] = []
        result_reduct[key].append(speed)
    result_reduct = [
        [setting, sum(speed_list) / len(speed_list)]
        for (setting, speed_list) in result_reduct.items()
    ]
    result_reduct = sorted(result_reduct, key=lambda item: -item[1])
    print(result_reduct)

    return result_reduct[0][0]



import torch
import cupy
from utils import qsgd_compress, qsgd_decompress
import argparse

# TODO: change to formal benchmark tasks https://github.com/ionelmc/pytest-benchmark
def main():
    parser = argparse.ArgumentParser(description="Test QSGD Compression Error")
    parser.add_argument(
        "--dimension", type=int, default=10000, metavar="N", help="length of tensor"
    )
    parser.add_argument(
        "--norms",
        type=str,
        default="2,4,8,16,inf",
        metavar="N",
        help="pnorm used for qsgd quantization",
    )
    parser.add_argument(
        "--bits",
        type=str,
        default="2,4,8",
        metavar="N",
        help="bits used for qsgd quantization",
    )
    parser.add_argument(
        "--random_test_num",
        type=int,
        default=100,
        metavar="N",
        help="number of test time",
    )
    args = parser.parse_args()

    norms = [float(i) for i in args.norms.split(",")]
    bits = [int(i) for i in args.bits.split(",")]
    print("norms: ", norms)
    print("bits: ", bits)
    errors = torch.zeros(len(norms), len(bits))
    random_test_num = args.random_test_num
    dimension = args.dimension

    torch.manual_seed(42)
    tensor = torch.rand([args.random_test_num, args.dimension]).cuda()
    cupy_stream = cupy.cuda.ExternalStream(torch.cuda.current_stream().cuda_stream)
    cupy_stream.use()
    for t in range(args.random_test_num):
        input_t = tensor[t, :].detach().clone()
        for i, porm in enumerate(norms):
            for j, bit in enumerate(bits):
                norm, packed_sign, compressed_ints = qsgd_compress(
                    input_t, pnorm=porm, quan_bits=bit
                )
                quan_t = qsgd_decompress(
                    norm, packed_sign, compressed_ints, quan_bits=bit
                )
                errors[i, j] += (
                    (quan_t - input_t).norm() / input_t.norm() / random_test_num
                )
    print("error: ", errors)


if __name__ == "__main__":
    main()



import torch
import copy
from bagua.torch_api.utils import collocate_params, flatten_module_params

__all__ = ["FusedOptimizer"]


class FusedOptimizer(torch.optim.Optimizer):
    """Convert any optimizer into a fused optimizer.

    This fused optimizer fuses multiple module parameter update kernel launches
    into one or a few, by flattening parameter tensors into one or more
    contiguous buckets.

    It can be used in conjunction with :func:`bagua.torch_api.bagua_init`. In this case,
    `Bagua` will do the fusions automatically, otherwise, you need to explicitly
    pass ``do_flatten=True``.

    Args:
        optimizer (torch.optim.Optimizer): Any PyTorch optimizer.
        do_flatten (bool): Whether to flatten the parameters. Default: ``False``.

    Returns:
        Fused optimizer.


    Example::
        To use in conjunction with :func:`bagua.torch_api.bagua_init`:

        >>> optimizer = torch.optim.Adadelta(model.parameters(), ....)
        >>> optimizer = bagua.torch_api.contrib.FusedOptimizer(optimizer)
        >>> model = model.with_bagua([optimizer], GradientAllReduceAlgorithm())

        To use alone or with :class:`torch.nn.parallel.DistributedDataParallel`, set `do_flatten` to be ``True``:

        >>> optimizer = torch.optim.Adadelta(model.parameters(), ....)
        >>> optimizer = bagua.torch_api.contrib.FusedOptimizer(optimizer, do_flatten=True)
    """

    def __init__(self, optimizer: torch.optim.Optimizer, do_flatten: bool = False):
        self.optimizer = copy.copy(optimizer)
        super(FusedOptimizer, self).__init__(optimizer.param_groups, optimizer.defaults)

        if do_flatten:
            f32_params = [
                param
                for group in self.optimizer.param_groups
                for param in group["params"]
                if param.type() == "torch.cuda.FloatTensor"
            ]
            f16_params = [
                param
                for group in self.optimizer.param_groups
                for param in group["params"]
                if param.type() == "torch.cuda.HalfTensor"
            ]

            flatten_module_params(f32_params, align_bytes=1)
            flatten_module_params(f16_params, align_bytes=1)

    def step(self, closure=None):
        r"""Performs a single optimization step (parameter update).

        Args:
            closure (callable): A closure that reevaluates the model and
                returns the loss. Optional for most optimizers.

        .. note::
            Unless otherwise specified, this function should not modify the
            ``.grad`` field of the parameters.
        """
        for group in self.optimizer.param_groups:
            params = group["params"]
            grouped_params = group_params_by_storage(params)

            fused_params = []

            for _, group_p in grouped_params.items():
                fused_params.extend(reorder_params(group_p))

            group["params"] = fused_params

        return self.optimizer.step(closure)


def reorder_params(params):
    """Input params share same storage, reorder them by their storage offset"""

    sorted_params = sorted(params, key=lambda x: x.storage_offset())

    grouped = []
    tmp_params = []

    for p in sorted_params:
        if len(tmp_params) > 0 and not is_contiguous_param(p, tmp_params[-1]):
            grouped.append(collocate_params(tmp_params))
            tmp_params = []

        tmp_params.append(p)

    if len(tmp_params) > 0:
        grouped.append(collocate_params(tmp_params))  # FIXME: potential OOM

    return grouped


def is_contiguous_param(a, b):
    allocate_size_a = (
        a.bagua_tensor.num_elem_allocated() if hasattr(a, "bagua_tensor") else a.numel()
    )
    allocate_size_b = (
        b.bagua_tensor.num_elem_allocated() if hasattr(b, "bagua_tensor") else b.numel()
    )
    return (
        a.data.storage_offset() == b.data.storage_offset() + allocate_size_b
        and a.grad.data.storage_offset()
        == b.grad.data.storage_offset() + allocate_size_b
    ) or (
        b.data.storage_offset() == a.data.storage_offset() + allocate_size_a
        and b.grad.data.storage_offset()
        == a.grad.data.storage_offset() + allocate_size_a
    )


def group_params_by_storage(params):
    grouped_params = {}
    for p in params:
        weight_storage = p.data.storage().data_ptr()
        param_list = grouped_params.get(weight_storage, [])
        param_list.append(p)
        grouped_params[weight_storage] = param_list

    return grouped_params



import unittest
import time
import numpy as np
from bagua.torch_api.utils import StatisticalAverage


class TestUtils(unittest.TestCase):
    def test_statistical_average(self):
        m = StatisticalAverage(
            last_update_time=time.time(),
            records=[5.0, 4.5005175, 3.5034241850204078],
            record_tail=(2.0166309999999985, 2.499061185570463),
        )
        time.sleep(1)

        m.record(6.0)

        test_answer_list = [
            (1.0, 6.0),
            (2.0, (6.0 + 5.0) / 2.0),
            (3.0, (6.0 + 4.5005175 * 2.0) / 3.0),
            (5.0, (6.0 + 3.5034241850204078 * 4.0) / 5.0),
            (7.0, (6.0 + 2.499061185570463 * (4.0 + 2.0166309999999985)) / 7.0),
        ]
        test_val_list = []
        for (last_n_seconds, _) in test_answer_list:
            test_val_list.append(m.get_records_mean(last_n_seconds))

        for a, (n, b) in zip(test_val_list, test_answer_list):
            self.assertTrue(
                np.isclose(a, b, atol=0.1),
                "last_n_seconds={}, test_val={}, answer={}, m={}".format(n, a, b, m),
            )

        self.assertTrue(
            np.isclose(
                m.total_recording_time(),
                4.0 + 2.0166309999999985 + 1,
                atol=0.1,
            ),
            "total_recording_time={}".format(m.total_recording_time()),
        )


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python3

from bagua.torch_api.bucket import BaguaBucket
from bagua.torch_api.tensor import BaguaTensor
from bagua.torch_api.distributed import BaguaModule
from bagua.torch_api.algorithms import Algorithm
from bagua.torch_api import get_world_size
from typing import List


class ByteGradAlgorithm(Algorithm):
    def __init__(self, average: bool = True):
        """
        Create an instance of the
        `ByteGrad <https://baguasys.github.io/tutorials/algorithms/bytegrad.html>`_
        algorithm.

        Args:
            average (bool): If True, the gradients on each worker are averaged.
                Otherwise, they are summed.
        """
        self.average = average

    def tensors_to_buckets(self, tensors: List[List[BaguaTensor]]) -> List[BaguaBucket]:
        """
        Given the bucketing suggestion from Bagua, return the actual Bagua buckets.
        The default implementation follows the suggestion to do the bucketing.

        Args:
            tensors: Bagua tensors grouped in different
                lists, representing Bagua's suggestion on how to bucketing the
                tensors.

        Returns:
            A list of Bagua buckets.
        """
        bagua_buckets = []
        for idx, bucket in enumerate(tensors):
            bagua_bucket = BaguaBucket(
                bucket, flatten=True, name=str(idx), alignment=get_world_size()
            )
            bagua_buckets.append(bagua_bucket)
        return bagua_buckets

    def init_operations(
        self,
        bagua_module: BaguaModule,
        bucket: BaguaBucket,
    ):
        bucket.clear_ops()
        bucket.append_centralized_synchronous_op(
            hierarchical=True,
            average=self.average,
            scattergather=True,
            compression="MinMaxUInt8",
        )



#!/usr/bin/env python3
"""
The Bagua communication library PyTorch interface.
"""
from .communication import (  # noqa: F401
    init_process_group,
    allreduce,
    broadcast,
)
from .distributed import BaguaModule  # noqa: F401
from .tensor import BaguaTensor  # noqa: F401
from .env import (  # noqa: F401
    get_rank,
    get_world_size,
    get_local_rank,
    get_local_size,
)
from . import contrib  # noqa: F401
from . import communication  # noqa: F401
from . import algorithms  # noqa: F401



#!/usr/bin/env python3

from bagua.torch_api.bucket import BaguaBucket
from bagua.torch_api.tensor import BaguaTensor
from bagua.torch_api.distributed import BaguaModule
from bagua.torch_api.algorithms import Algorithm
from typing import List
import torch


class DecentralizedAlgorithm(Algorithm):
    def __init__(
        self,
        peer_selection_mode: str = "all",
        compression: str = None,
        communication_interval: int = 1,
    ):
        """
        Create an instance of the
        `Decentralized <https://baguasys.github.io/tutorials/algorithms/decentralized.html>`_
        algorithm.

        Args:
            peer_selection_mode (str): Can be "all" or "shift_one". "all" means all workers'
                weights are averaged in each communication step. "shift_one" means each worker
                selects a different peer to do weights average in each communication step.
            compression (str): Not supported yet.
            communication_interval (int): Number of iterations between two communication steps.
        """
        self.peer_selection_mode = peer_selection_mode
        self.compression = compression
        self.communication_interval = communication_interval
        self.tensor_groups = []

    def init_tensors(self, bagua_module: BaguaModule) -> List[BaguaTensor]:
        parameters = bagua_module.bagua_build_params()
        self.tensors = [
            param.ensure_bagua_tensor(name, bagua_module.bagua_module_name)
            for name, param in parameters.__reversed__()
        ]
        return self.tensors

    def init_forward_pre_hook(self, bagua_module: BaguaModule):
        def hook(input):
            for tensor in self.tensors:
                tensor.bagua_mark_communication_ready()

        return hook

    def init_backward_hook(self, bagua_module: BaguaModule):
        def hook(parameter_name, parameter):
            return

        return hook

    def init_post_backward_hook(self, bagua_module: BaguaModule):
        def hook():
            bagua_module._bagua_backend.wait_pending_comm_ops()
            torch.cuda.synchronize()
            bagua_module._bagua_backend.execute_post_backward_comm_ops()
            bagua_module._bagua_backend.wait_pending_post_backward_comm_ops()

        return hook

    def init_operations(
        self,
        bagua_module: BaguaModule,
        bucket: BaguaBucket,
    ):
        bucket.clear_ops()
        bucket.append_decentralized_synchronous_op(
            hierarchical=True,
            peer_selection_mode=self.peer_selection_mode,
            communication_interval=self.communication_interval,
        )



#!/usr/bin/env python3
from bagua.torch_api.bucket import BaguaBucket
from bagua.torch_api.tensor import BaguaTensor
from bagua.torch_api import get_world_size
from bagua.torch_api.distributed import BaguaModule
from bagua.torch_api.algorithms import Algorithm
from torch.optim.optimizer import Optimizer
import torch
import math
from typing import List, Tuple


class QAdamOptimizer(Optimizer):
    def __init__(
        self,
        params,
        lr: float = 1e-3,
        warmup_steps: int = 100,
        betas: Tuple[float, float] = (0.9, 0.999),
        eps: float = 1e-8,
        weight_decay: float = 0.0,
    ):
        """
        Create a dedicated optimizer used for QAdam algorithm.

        Args:
            params (iterable): iterable of parameters to optimize or dicts defining
                parameter groups
            lr: learning rate
            warmup_steps: number of steps to do warm up in the begining of training.
            betas: coefficients used for computing running averages of gradient and its square
            eps: term added to the denominator to improve numerical stability
            weight_decay: weight decay (L2 penalty)
        """
        if not 0.0 <= lr:
            raise ValueError("Invalid learning rate: {}".format(lr))
        if not 0.0 <= eps:
            raise ValueError("Invalid epsilon value: {}".format(eps))
        if not 0.0 <= betas[0] < 1.0:
            raise ValueError("Invalid beta parameter at index 0: {}".format(betas[0]))
        if not 0.0 <= betas[1] < 1.0:
            raise ValueError("Invalid beta parameter at index 1: {}".format(betas[1]))
        defaults = dict(lr=lr, betas=betas, eps=eps, weight_decay=weight_decay)
        super(QAdamOptimizer, self).__init__(params, defaults)

        self.params_in_group = []
        self.exp_avgs_in_group = []
        self.step_id = 0
        self.warmup_steps = warmup_steps

        # initialize momentum and variance
        for group_id, group in enumerate(self.param_groups):
            params_with_grad = []
            exp_avgs = []
            for p in group["params"]:
                params_with_grad.append(p)
                state = self.state[p]
                if len(state) == 0:
                    state["exp_avg"] = torch.zeros_like(
                        p, memory_format=torch.preserve_format
                    )
                    state["exp_avg_sq"] = torch.zeros_like(
                        p, memory_format=torch.preserve_format
                    )
                exp_avgs.append(state["exp_avg"])
            self.params_in_group.append(params_with_grad)
            self.exp_avgs_in_group.append(exp_avgs)

    def __setstate__(self, state):
        super(QAdamOptimizer, self).__setstate__(state)

    def step(self, closure=None):
        self.step_id += 1
        for group_id, group in enumerate(self.param_groups):

            lr = group["lr"]
            weight_decay = group["weight_decay"]
            beta1, beta2 = group["betas"]
            eps = group["eps"]

            for param_id, param in enumerate(group["params"]):
                state = self.state[param]

                if self.step_id < self.warmup_steps:
                    state["exp_avg"].mul_(beta1).add_(param.grad, alpha=1 - beta1)
                    state["exp_avg_sq"].mul_(beta2).addcmul_(
                        param.grad, param.grad, value=1 - beta2
                    )

                bias_correction1 = 1 - beta1 ** self.step_id
                bias_correction2 = 1 - beta2 ** self.step_id

                denom = (state["exp_avg_sq"].sqrt() / math.sqrt(bias_correction2)).add_(
                    eps
                )
                step_size = lr / bias_correction1
                update = state["exp_avg"] / denom
                param.data.add_(-step_size * update)


class QAdamAlgorithm(Algorithm):
    def __init__(self, q_adam_optimizer: QAdamOptimizer, hierarchical: bool = True):
        """
        Create an instance of the
        `QAdam Algorithm <https://baguasys.github.io/tutorials/algorithms/q-adam.html>`_
        .

        Args:
            q_adam_optimizer: A QAdamOptimizer initialized with model parameters.
            hierarchical: Enable hierarchical communication.
        """
        self.hierarchical = hierarchical
        self.optimizer = q_adam_optimizer
        self.warmup_steps = self.optimizer.warmup_steps

    def need_reset(self):
        if self.optimizer.step_id == self.warmup_steps:
            print(
                "QAdam starts to compress from step {}".format(self.optimizer.step_id)
            )
            return True
        else:
            return False

    def init_tensors(self, bagua_module: BaguaModule):
        parameters = bagua_module.bagua_build_params()

        for idx, (name, param) in enumerate(parameters.__reversed__()):
            param._q_adam_name = name
            param._q_adam_idx = idx

        tensor_groups = []
        for param_group, m_group in zip(
            self.optimizer.params_in_group, self.optimizer.exp_avgs_in_group
        ):
            for param, exp_avgs in zip(param_group, m_group):
                if self.optimizer.step_id < self.warmup_steps:
                    registered_tensor = param.bagua_ensure_grad().ensure_bagua_tensor(
                        param._q_adam_name, bagua_module.bagua_module_name
                    )
                    param._q_adam_grad = registered_tensor
                    registered_tensor._q_adam_idx = param._q_adam_idx
                else:
                    registered_tensor = exp_avgs.ensure_bagua_tensor(
                        param._q_adam_name, bagua_module.bagua_module_name
                    )
                    registered_tensor._q_adam_grad = param.bagua_ensure_grad()
                    param._q_adam_momentum = registered_tensor
                    registered_tensor._q_adam_idx = param._q_adam_idx
                tensor_groups.append(registered_tensor)
        tensor_groups.sort(key=lambda x: x._q_adam_idx)
        return tensor_groups

    def tensors_to_buckets(self, tensors: List[List[BaguaTensor]]) -> List[BaguaBucket]:
        bagua_buckets = []
        for idx, bucket in enumerate(tensors):
            bagua_bucket = BaguaBucket(
                bucket, flatten=True, name=str(idx), alignment=get_world_size()
            )
            bagua_buckets.append(bagua_bucket)
        return bagua_buckets

    def init_operations(
        self,
        bagua_module: BaguaModule,
        bucket: BaguaBucket,
    ):
        bucket.clear_ops()
        if self.optimizer.step_id < self.warmup_steps:
            bucket.append_centralized_synchronous_op(
                hierarchical=False,
                average=True,
            )
        else:

            def calculate_momentum(*args):
                beta1, beta2 = self.optimizer.param_groups[0]["betas"]
                for tensor in bucket.tensors:
                    tensor.mul_(beta1).add_(tensor._q_adam_grad, alpha=1 - beta1)

            bucket.append_python_op(calculate_momentum)
            bucket.append_centralized_synchronous_op(
                hierarchical=self.hierarchical,
                average=True,
                scattergather=True,
                compression="MinMaxUInt8",
            )

    def init_backward_hook(self, bagua_module: BaguaModule):
        def hook_momentum(parameter_name, parameter):
            parameter._q_adam_momentum.bagua_mark_communication_ready()

        def hook_grad(parameter_name, parameter):
            assert (
                parameter.grad.data_ptr() == parameter._q_adam_grad.data_ptr()
            ), "gradient data_ptr should match _q_adam_grad data_ptr"
            parameter._q_adam_grad.bagua_mark_communication_ready()

        return (
            hook_grad if self.optimizer.step_id < self.warmup_steps else hook_momentum
        )



import torch
import math
import torch.distributed as dist
from torch.utils.data.sampler import Sampler
from torch.utils.data.dataset import Dataset
from typing import Optional, Iterator, Callable
from collections import OrderedDict

__all__ = ["LoadBalancingDistributedSampler", "LoadBalancingDistributedBatchSampler"]


class LoadBalancingDistributedSampler(Sampler):
    r"""Sampler that restricts data loading to a subset of the dataset.

    This sampler use a `complexity_fn` to calculate each sample's computational
    complexity and make each batch get similar computation complexity.

    This is useful in scenarios like speech and NLP, where each batch has variable
    length and distributed training suffers from straggler problem.

    The usage is similar to :class:`torch.utils.data.DistributedSampler`, where each
    process loads a subset of the original dataset that is exclusive to it.

    .. note::
        Dataset is assumed to be of constant size.

    Args:
        dataset: Dataset used for sampling.
        complexity_fn(Callable): A function whose input is a sample and output is an integer as a
            measure of the computational complexity of the sample.
        num_replicas (int, optional): Number of processes participating in
            distributed training. By default, :attr:`world_size` is retrieved from the
            current distributed group.
        rank (int, optional): Rank of the current process within :attr:`num_replicas`.
            By default, :attr:`rank` is retrieved from the current distributed
            group.
        shuffle (bool, optional): If ``True`` (default), sampler will shuffle the
            indices.
        seed (int, optional): random seed used to shuffle the sampler if
            :attr:`shuffle=True`. This number should be identical across all
            processes in the distributed group. Default: ``0``.
        drop_last (bool, optional): if ``True``, then the sampler will drop the
            tail of the data to make it evenly divisible across the number of
            replicas. If ``False``, the sampler will add extra indices to make
            the data evenly divisible across the replicas. Default: ``False``.
        random_level (float, optional): A float varies from ``0`` and ``1`` that controls the extent
            of load balance. ``0`` means the best load balance, while ``1`` means the opposite.

    .. warning::
        In distributed mode, calling the :meth:`set_epoch` method at
        the beginning of each epoch **before** creating the :class:`DataLoader` iterator
        is necessary to make shuffling work properly across multiple epochs. Otherwise,
        the same ordering will be always used.

    Example::
        Define your `complexity_fn`, which accepts a dataset sample as its input and produces an integer
        as the sample's computational complexity.

        >>> dataset = torch.utils.data.TensorDataset(torch.randn(n, 2), torch.randperm(n))
        >>> complexity_fn = lambda x: x[1]

        Below is the usage of :class:`LoadBalancingDistributedSampler` and :class:`DataLoader`:

        >>> sampler = bagua.torch_api.contrib.LoadBalancingDistributedSampler(
        ...     dataset,
        ...     complexity_fn=complexity_fn) if is_distributed else None
        >>> loader = torch.utils.data.DataLoader(dataset,
        ...     shuffle=(sampler is None),
        ...     sampler=sampler)
        >>>
        >>> for epoch in range(start_epoch, n_epochs):
        ...     if is_distributed:
        ...         sampler.set_epoch(epoch)
        ...     train(loader)
    """

    def __init__(
        self,
        dataset: Dataset,
        complexity_fn: Callable[..., int],
        num_replicas: Optional[int] = None,
        rank: Optional[int] = None,
        shuffle: bool = True,
        seed: int = 0,
        drop_last: bool = False,
        random_level: float = 0,
    ) -> None:
        if num_replicas is None:
            if not dist.is_available():
                raise RuntimeError("Requires distributed package to be available")
            num_replicas = dist.get_world_size()
        if rank is None:
            if not dist.is_available():
                raise RuntimeError("Requires distributed package to be available")
            rank = dist.get_rank()
        if rank >= num_replicas or rank < 0:
            raise ValueError(
                "Invalid rank {}, rank should be in the interval"
                " [0, {}]".format(rank, num_replicas - 1)
            )
        self.dataset = dataset
        self.num_replicas = num_replicas
        self.rank = rank
        self.epoch = 0
        self.drop_last = drop_last

        # If the dataset length is evenly divisible by # of replicas, then there
        # is no need to drop any data, since the dataset will be split equally.
        dataset_len = len(self.dataset)  # type: ignore
        if self.drop_last and dataset_len % self.num_replicas != 0:  # type: ignore
            # Split to nearest available length that is evenly divisible.
            # This is to ensure each rank receives the same amount of data when
            # using this Sampler.
            self.num_samples = math.ceil(
                # `type:ignore` is required because Dataset cannot provide a default __len__
                # see NOTE in pytorch/torch/utils/data/sampler.py
                (dataset_len - self.num_replicas)
                / self.num_replicas
            )
        else:
            self.num_samples = math.ceil(dataset_len / self.num_replicas)  # type: ignore
        self.total_size = self.num_samples * self.num_replicas
        self.shuffle = shuffle
        self.seed = seed

        self.item_complexity_map = dict()
        for item_index in range(dataset_len):
            self.item_complexity_map[item_index] = complexity_fn(
                self.dataset[item_index]
            )

        self.ordered_item_complexity_map = OrderedDict(
            sorted(self.item_complexity_map.items(), key=lambda t: t[1])
        )
        max_complexity = max(self.item_complexity_map.values())
        min_complexity = min(self.item_complexity_map.values())

        if random_level < 0.0 or random_level > 1.0:
            raise ValueError(
                "Invalid random level {}, shoule be in the range [0.0, 1.0]".format(
                    random_level
                )
            )

        self.random_number = int((max_complexity - min_complexity) * random_level + 1)

    def shuffle_chunks(self):
        def chunks_wrap_padding(lst, n):
            """Yield successive n-sized chunks from lst."""
            num_chunks = max(1, self.num_samples)
            num_elements = num_chunks * n
            current_lst = []
            for i in range(num_elements):
                current_lst.append(lst[i % len(lst)])
                if len(current_lst) == n:
                    yield current_lst
                    current_lst = []

        if self.shuffle:
            # deterministically shuffle based on epoch and seed
            g = torch.Generator()
            g.manual_seed(self.seed + self.epoch)

            if self.random_number > 0:
                item_complexity_map = self.item_complexity_map.copy()
                complexity_random_ints = torch.randint(
                    self.random_number, (len(item_complexity_map),), generator=g
                ).tolist()

                for k, v in zip(item_complexity_map, complexity_random_ints):
                    item_complexity_map[k] += v

                ordered_item_complexity_map = OrderedDict(
                    sorted(item_complexity_map.items(), key=lambda t: t[1])
                )
            else:
                ordered_item_complexity_map = self.ordered_item_complexity_map

            index_chunks = list(
                chunks_wrap_padding(
                    list(ordered_item_complexity_map.keys()), self.num_replicas
                )
            )

            chunk_indices = torch.randperm(len(index_chunks), generator=g).tolist()  # type: ignore
        else:
            index_chunks = list(
                chunks_wrap_padding(
                    list(self.ordered_item_complexity_map.keys()), self.num_replicas
                )
            )
            chunk_indices = list(range(len(index_chunks)))  # type: ignore

        if not self.drop_last:
            # add extra samples to make it evenly divisible
            padding_size = self.num_samples - len(chunk_indices)
            if padding_size <= len(chunk_indices):
                chunk_indices += chunk_indices[:padding_size]
            else:
                chunk_indices += (
                    chunk_indices * math.ceil(padding_size / len(chunk_indices))
                )[:padding_size]
        else:
            # remove tail of data to make it evenly divisible.
            chunk_indices = chunk_indices[: self.num_samples]
        assert len(chunk_indices) == self.num_samples
        return index_chunks, chunk_indices

    def __iter__(self) -> Iterator:
        index_chunks, chunk_indices = self.shuffle_chunks()
        # subsample
        indices = [index_chunks[i][self.rank] for i in chunk_indices]
        assert len(indices) == self.num_samples

        return iter(indices)

    def __len__(self) -> int:
        return self.num_samples

    def set_epoch(self, epoch: int) -> None:
        r"""
        Sets the epoch for this sampler. When :attr:`shuffle=True`, this ensures all replicas
        use a different random ordering for each epoch. Otherwise, the next iteration of this
        sampler will yield the same ordering.

        Args:
            epoch (int): Epoch number.
        """
        self.epoch = epoch


class LoadBalancingDistributedBatchSampler(Sampler):
    r"""Wraps another load balance sampler to yield variable sized mini-batches.

    Args:
        sampler (LoadBalancingDistributedSampler): Load balance sampler.
        batch_fn (Callable): Callable to yield mini-batch indices.
        drop_last (bool): If ``True``, the sampler will drop the last few batches exceeding
            the least number of batches among replicas, otherwises, the number of batches
            on each replica will be padded to the same.


    `batch_fn` will have the signature of
    ``def batch_fn(indices: List[int]) -> List[List[int]]``.

    Example::

        >>> from bagua.torch_api.contrib import LoadBalancingDistributedSampler, \
        ...     LoadBalancingDistributedBatchSampler
        >>>
        >>> sampler = LoadBalancingDistributedSampler(dataset, complexity_fn=complexity_fn)
        >>> batch_sampler = LoadBalancingDistributedBatchSampler(sampler, batch_fn=batch_fn)
        >>> loader = torch.utils.data.DataLoader(dataset, batch_sampler=batch_sampler)
        >>>
        >>> for epoch in range(start_epoch, n_epochs):
        ...     batch_sampler.set_epoch(epoch)
        ...     train(loader)

    """

    def __init__(
        self,
        sampler: LoadBalancingDistributedSampler,
        batch_fn,
        drop_last: bool = False,
    ) -> None:
        if not isinstance(sampler, LoadBalancingDistributedSampler):
            raise ValueError(
                "sampler should be of LoadBalancingDistributedSampler type."
            )

        if sampler.drop_last:
            raise ValueError("drop_last of sampler should be False")

        self.sampler = sampler
        self.batch_fn = batch_fn
        self.drop_last = drop_last

        self.num_replicas = self.sampler.num_replicas
        self.rank = self.sampler.rank

        self.generate_batches()

    def generate_batches(self):
        index_chunks, chunk_indices = self.sampler.shuffle_chunks()

        batches = []
        for rank in range(self.num_replicas):
            sub_indices = [index_chunks[i][rank] for i in chunk_indices]
            batches.append(self.batch_fn(sub_indices))

        self.total_batch = (
            max([len(b) for b in batches])
            if not self.drop_last
            else min([len(b) for b in batches])
        )

        # here {len(batches[self.rank]) - self.total_batch} batches dropped for
        # rank {self.rank}
        if self.total_batch < len(batches[self.rank]):
            pass

        self.padded_batches = [
            batch + batch[: self.total_batch - len(batch)] for batch in batches
        ]

    def __iter__(self):
        return iter(self.padded_batches[self.rank])

    def __len__(self):
        return self.total_batch

    def set_epoch(self, epoch: int) -> None:
        r"""
        Sets the epoch for this sampler. When :attr:`shuffle=True`, this ensures all replicas
        use a different random ordering for each epoch. Otherwise, the next iteration of this
        sampler will yield the same ordering.

        Args:
            epoch (int): Epoch number.
        """
        self.sampler.set_epoch(epoch)
        self.generate_batches()



r"""
This is a script forked from pytorch, use reference https://github.com/pytorch/pytorch/blob/master/torch/distributed/launch.py
"""

import time
import signal
import sys
import subprocess
import os
from argparse import ArgumentParser, REMAINDER
from typing import Optional, IO, List, Any
import logging

node_local_rank_stdout_filename = "node_{}_local_rank_{}_stdout"
node_local_rank_stderr_filename = "node_{}_local_rank_{}_stderr"


def parse_args():
    """
    Helper function parsing the command line options
    @retval ArgumentParser
    """
    parser = ArgumentParser(
        description="PyTorch distributed training launch "
        "helper utility that will spawn up "
        "multiple distributed processes"
    )

    # Optional arguments for the launch helper
    parser.add_argument(
        "--nnodes",
        type=int,
        default=1,
        help="The number of nodes to use for distributed " "training",
    )
    parser.add_argument(
        "--node_rank",
        type=int,
        default=0,
        help="The rank of the node for multi-node distributed " "training",
    )
    parser.add_argument(
        "--nproc_per_node",
        type=int,
        default=1,
        help="The number of processes to launch on each node, "
        "for GPU training, this is recommended to be set "
        "to the number of GPUs in your system so that "
        "each process can be bound to a single GPU.",
    )
    parser.add_argument(
        "--master_addr",
        default="127.0.0.1",
        type=str,
        help="Master node (rank 0)'s address, should be either "
        "the IP address or the hostname of node 0, for "
        "single node multi-proc training, the "
        "--master_addr can simply be 127.0.0.1",
    )
    parser.add_argument(
        "--master_port",
        default=29500,
        type=int,
        help="Master node (rank 0)'s free port that needs to "
        "be used for communication during distributed "
        "training",
    )
    parser.add_argument(
        "--bagua_service_port",
        default=29501,
        type=int,
    )
    parser.add_argument(
        "--set_additional_flag",
        default=False,
        action="store_true",
        help="Pass --local_rank as argument.",
    )
    parser.add_argument(
        "-m",
        "--module",
        default=False,
        action="store_true",
        help="Changes each process to interpret the launch script "
        "as a python module, executing with the same behavior as"
        "'python -m'.",
    )
    parser.add_argument(
        "--no_python",
        default=False,
        action="store_true",
        help='Do not prepend the training script with "python" - just exec '
        "it directly. Useful when the script is not a Python script.",
    )
    parser.add_argument(
        "--logdir",
        default=None,
        type=str,
        help=f"""Relative path to write subprocess logs to. Passing in a relative
        path will create a directory if needed, and write the stdout and stderr to files
        {node_local_rank_stdout_filename} and {node_local_rank_stderr_filename}. Note that
        successive runs with the  same path to write logs to will overwrite existing logs,
        so be sure to save logs as needed.""",
    )

    # AutoTune args
    parser.add_argument(
        "--autotune_level",
        type=int,
        default=0,
        help="Bagua automatic super parameter search level. The higher the level, the better the theoretical effect, and the longer it takes",
    )
    parser.add_argument(
        "--is_output_autotune_log",
        action="store_true",
        default=False,
        help="Whether autotune output log or not. default is False",
    )
    parser.add_argument(
        "--report_metrics",
        action="store_true",
        help="switch of bagua report metrics function",
    )
    parser.add_argument("--autotune_max_samples", type=int, default=60)
    parser.add_argument("--autotune_sampling_confidence_time", type=float, default=5.0)
    parser.add_argument("--autotune_warmup_time", type=float, default=30.0)
    parser.add_argument(
        "--default_bucket_size", type=int, default=10 * 1024 ** 2
    )  # noqa: E501

    parser.add_argument("--host_list", type=str)
    parser.add_argument("--ssh_port", type=int)

    # positional
    parser.add_argument(
        "training_script",
        type=str,
        help="The full path to the single GPU training "
        "program/script to be launched in parallel, "
        "followed by all the arguments for the "
        "training script",
    )

    # rest from the training program
    parser.add_argument("training_script_args", nargs=REMAINDER)
    return parser.parse_args()


def set_bagua_env(args, current_env):
    current_env["BAGUA_SERVICE_PORT"] = str(args.bagua_service_port)
    current_env["BAGUA_DEFAULT_BUCKET_SIZE"] = str(args.default_bucket_size)
    current_env["BAGUA_AUTOTUNE"] = str(args.autotune_level)
    current_env["BAGUA_AUTOTUNE_MAX_SAMPLES"] = str(args.autotune_max_samples)
    current_env["BAGUA_AUTOTUNE_SAMPLING_CONFIDENCE_TIME_S"] = str(
        args.autotune_sampling_confidence_time
    )
    current_env["BAGUA_AUTOTUNE_WARMUP_TIME_S"] = str(args.autotune_warmup_time)
    current_env["BAGUA_IS_OUTPUT_AUTOTUNE_LOG"] = str(int(args.is_output_autotune_log))


def main():
    logging.basicConfig(
        format="%(asctime)s,%(msecs)d %(levelname)-8s [%(filename)s:%(lineno)d] %(message)s",
        datefmt="%Y-%m-%d:%H:%M:%S",
        level=logging.INFO,
    )

    args = parse_args()

    # world size in terms of number of processes
    dist_world_size = args.nproc_per_node * args.nnodes

    # set PyTorch distributed related environmental variables
    current_env = os.environ.copy()
    current_env["MASTER_ADDR"] = args.master_addr
    current_env["MASTER_PORT"] = str(args.master_port)
    current_env["WORLD_SIZE"] = str(dist_world_size)

    set_bagua_env(args, current_env)

    processes: List[Any] = []

    if "OMP_NUM_THREADS" not in os.environ and args.nproc_per_node > 1:
        current_env["OMP_NUM_THREADS"] = str(1)
        print(
            "*****************************************\n"
            "Setting OMP_NUM_THREADS environment variable for each process "
            "to be {} in default, to avoid your system being overloaded, "
            "please further tune the variable for optimal performance in "
            "your application as needed. \n"
            "*****************************************".format(
                current_env["OMP_NUM_THREADS"]
            )
        )

    if args.logdir:
        # Possibly create the directory to write subprocess log output to.
        if os.path.exists(args.logdir):
            if not os.path.isdir(args.logdir):
                raise ValueError("argument --logdir must be a path to a directory.")
        else:
            # create the relative directory
            os.mkdir(os.path.join(os.getcwd(), args.logdir))

    subprocess_file_handles = []

    for local_rank in range(0, args.nproc_per_node):
        # each process's rank
        dist_rank = args.nproc_per_node * args.node_rank + local_rank
        current_env["RANK"] = str(dist_rank)
        current_env["LOCAL_RANK"] = str(local_rank)
        current_env["LOCAL_WORLD_SIZE"] = str(args.nproc_per_node)

        # spawn the processes
        with_python = not args.no_python
        cmd = []
        if with_python:
            cmd = [sys.executable, "-u"]
            if args.module:
                cmd.append("-m")
        else:
            if args.module:
                raise ValueError(
                    "Don't use both the '--no_python' flag and the '--module' flag at the same time."
                )

        cmd.append(args.training_script)

        if args.set_additional_flag:
            cmd.append("--local_rank={}".format(local_rank))

        cmd.extend(args.training_script_args)

        stdout_handle: Optional[IO]
        stderr_handle: Optional[IO]
        if args.logdir:
            directory_path = os.path.join(os.getcwd(), args.logdir)
            node_rank = args.node_rank
            stdout_file_name = node_local_rank_stdout_filename.format(
                node_rank, local_rank
            )
            stderr_file_name = node_local_rank_stderr_filename.format(
                node_rank, local_rank
            )
            stdout_handle = open(os.path.join(directory_path, stdout_file_name), "w")
            stderr_handle = open(os.path.join(directory_path, stderr_file_name), "w")
            subprocess_file_handles.append((stdout_handle, stderr_handle))
            stdout_name = stdout_handle.name
            stderr_name = stderr_handle.name
            print(
                f"""Note: Stdout and stderr for node {node_rank} rank {local_rank} will
            be written to {stdout_name}, {stderr_name} respectively."""
            )

        sig_names = {2: "SIGINT", 15: "SIGTERM"}
        last_return_code = None

        def sigkill_handler(signum, frame):
            for process in processes:
                print(f"Killing subprocess {process.pid}")
                try:
                    process.kill()
                except Exception:
                    pass
            if last_return_code is not None:
                raise subprocess.CalledProcessError(
                    returncode=last_return_code, cmd=cmd
                )
            if signum in sig_names:
                print(f"Main process received {sig_names[signum]}, exiting")
            sys.exit(1)

        # pass SIGINT/SIGTERM to children if the parent is being terminated
        signal.signal(signal.SIGINT, sigkill_handler)
        signal.signal(signal.SIGTERM, sigkill_handler)

        stdout_handle = (
            None
            if not subprocess_file_handles
            else subprocess_file_handles[local_rank][0]
        )
        stderr_handle = (
            None
            if not subprocess_file_handles
            else subprocess_file_handles[local_rank][1]
        )
        process = subprocess.Popen(
            cmd, env=current_env, stdout=stdout_handle, stderr=stderr_handle
        )
        processes.append(process)

    try:
        alive_processes = set(processes)
        while len(alive_processes):
            finished_processes = []
            for process in alive_processes:
                if process.poll() is None:
                    # the process is still running
                    continue
                else:
                    if process.returncode != 0:
                        last_return_code = process.returncode  # for sigkill_handler
                        sigkill_handler(signal.SIGTERM, None)  # not coming back
                    else:
                        # exited cleanly
                        finished_processes.append(process)
            alive_processes = set(alive_processes) - set(finished_processes)

            time.sleep(1)
    finally:
        # close open file descriptors
        for (stdout_handle, stderr_handle) in subprocess_file_handles:
            stdout_handle.close()
            stderr_handle.close()


if __name__ == "__main__":
    main()



#!/usr/bin/env python3

from __future__ import annotations
from bagua.torch_api.communication import get_backend
from typing import List, Callable, Optional

import bagua_core as B
import torch

from bagua.torch_api.tensor import BaguaTensor
from bagua.torch_api.utils import check_contiguous


class BaguaBucket:
    def __init__(
        self, tensors: List[BaguaTensor], name: str, flatten: bool, alignment: int = 1
    ) -> None:
        """
        Create a Bagua bucket with a list of Bagua tensors.

        Args:
            tensors: A list of Bagua tensors to be put in the
                bucket.
            name: The unique name of the bucket.
            flatten: If True, flatten the input tensors so that they are
                contiguous in memory.
            alignment: If alignment > 1, Bagua will create a padding tensor to
                the bucket so that the total number of elements in the bucket divides
                the given alignment.
        """
        self.tensors = tensors
        """
        The tensors contained within the bucket.
        """
        self.bagua_module_name = tensors[0].bagua_module_name
        for tensor in self.tensors:
            assert (
                self.bagua_module_name == tensor.bagua_module_name
            ), "every tensor in the same bucket should have the same model name"
        self._bagua_backend = get_backend(self.bagua_module_name)
        self.name = name
        """
        The bucket's name.
        """
        self.padding_tensor = None

        if alignment > 1:
            padding = sum(tensor.numel() for tensor in self.tensors) % alignment
            if padding > 0:
                padding = alignment - padding
                # padding tensor must be of name bagua_padding_tensor, so that they are always marked as ready for communication in the backend
                self.padding_tensor = torch.zeros(
                    padding, dtype=self.tensors[0].dtype, device=self.tensors[0].device
                ).to_bagua_tensor("bagua_padding_tensor_bucket_" + name)

        self._all_tensors = (
            self.tensors + [self.padding_tensor]
            if self.padding_tensor is not None
            else self.tensors
        )

        self.backend_tensor = None
        self.flatten = flatten
        if self.flatten:
            self._flatten_()

        self.backend_bucket = B.BaguaBucketPy(
            name, [tensor._bagua_backend_tensor for tensor in self._all_tensors]
        )

        for tensor in self._all_tensors:
            tensor._bagua_bucket = self

    def _flatten_(self):
        """
        Flatten inner tensors in place.
        """
        if self.check_flatten():
            return

        if len(self._all_tensors) == 0:
            return
        total_size = 0
        for tensor in self._all_tensors:
            total_size += tensor.numel()

        flatten_tensor = torch.zeros(total_size, dtype=self._all_tensors[0].dtype).to(
            self._all_tensors[0].device
        )
        flatten_storage = flatten_tensor.storage()

        offset = 0
        for tensor in self._all_tensors:
            # copy data
            flatten_tensor[offset : offset + tensor.numel()] = tensor.data.reshape(-1)
            tensor.bagua_set_storage(flatten_storage, offset)
            offset += tensor.numel()
        # check
        assert self.check_flatten()

    def check_flatten(self) -> bool:
        """
        Returns:
            True if the bucket's tensors are contiguous in memory.
        """
        return check_contiguous(self._all_tensors)

    def append_python_op(self, python_function: Callable[[str], None]) -> BaguaBucket:
        """
        Append a Python operation to a bucket. A Python operation is a Python function that
        takes the bucket's name and returns ``None``. It can do arbitrary things within the
        function body.

        The operations will be executed by the Bagua backend in the order they are appended
        when all the tensors within the bucket are marked ready.

        Args:
            python_function: The Python operation function.

        Returns:
            The bucket itself.
        """

        def wrapper_function_factory(pyop):
            def wrapped_pyop(name):
                with torch.cuda.stream(self._bagua_backend.stream):
                    return pyop(name)

            return wrapped_pyop

        self.backend_bucket.append_python_op(wrapper_function_factory(python_function))
        return self

    def append_centralized_synchronous_op(
        self,
        hierarchical: bool = False,
        average: bool = True,
        scattergather: bool = False,
        compression: Optional[str] = None,
    ) -> BaguaBucket:
        """
        Append a centralized synchronous operation to a bucket. It will sum or average the tensors in the bucket
        for all workers.

        The operations will be executed by the Bagua backend in the order they are appended
        when all the tensors within the bucket are marked ready.

        Args:
            hierarchical (bool): Enable hierarchical communication. Which means the GPUs on the same machine
                will communicate will each other first. After that, machines do inter-node communication. This can
                boost performance when the inter-node communication cost is high.
            average (bool): If True, the gradients on each worker are averaged. Otherwise, they are summed.
            scattergather (bool): If true, the communication between workers are done with scatter gather instead
                of allreduce. This is required for using compression.
            compression: If not None, the tensors will be compressed for communication. Currently "MinMaxUInt8" is
                supported.
        Returns:
            The bucket itself.
        """
        if hierarchical:
            self.backend_bucket.append_centralized_synchronous_op(
                self._bagua_backend.internode_communicator,
                self._bagua_backend.intranode_communicator,
                hierarchical=hierarchical,
                average=average,
                scattergather=scattergather,
                compression=compression,
            )
        else:
            self.backend_bucket.append_centralized_synchronous_op(
                self._bagua_backend.global_communicator,
                None,
                hierarchical=hierarchical,
                average=average,
                scattergather=scattergather,
                compression=compression,
            )
        return self

    def append_decentralized_synchronous_op(
        self,
        hierarchical: bool = True,
        peer_selection_mode: str = "all",
        communication_interval: int = 1,
    ) -> BaguaBucket:
        """
        Append a decentralized synchronous operation to a bucket. It will do gossipy style model averaging among workers.

        The operations will be executed by the Bagua backend in the order they are appended
        when all the tensors within the bucket are marked ready.

        Args:
            hierarchical (bool): Enable hierarchical communication. Which means the GPUs on the same machine
                will communicate will each other first. After that, machines do inter-node communication. This can
                boost performance when the inter-node communication cost is high.
            peer_selection_mode (str): Can be "all" or "shift_one". "all" means all workers'
                weights are averaged in each communication step. "shift_one" means each worker
                selects a different peer to do weights average in each communication step.
            communication_interval (int): Number of iterations between two communication steps.
        Returns:
            The bucket itself.
        """
        self.backend_bucket.append_decentralized_synchronous_op(
            self._bagua_backend.internode_communicator,
            self._bagua_backend.intranode_communicator,
            hierarchical=hierarchical,
            compression=None,
            peer_selection_mode=peer_selection_mode,
            communication_interval=communication_interval,
        )
        return self

    def clear_ops(self) -> BaguaBucket:
        """
        Clear the previously appended operations.
        """
        self.backend_bucket.clear_ops()
        return self

    def bytes(self) -> int:
        """Returns the total number of bytes occupied by the bucket.

        Returns:
            int: number of bucket bytes
        """
        return sum(tensor.numel() * tensor.element_size() for tensor in self.tensors)



import copy
import csv
import requests
import os
import time
import threading
import tempfile
import json
import logging
import collections
import math
import multiprocessing
from flask import Flask, request
from .autotune import BayesianOptimizer, IntParam, BoolParam
from bagua.bagua_define import (
    TensorDtype,
    TensorDeclaration,
    BaguaHyperparameter,
)
import numpy as np
from typing import Dict, List, Tuple
from collections import OrderedDict


class NpEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, np.integer):
            return int(obj)
        elif isinstance(obj, np.floating):
            return float(obj)
        elif isinstance(obj, np.ndarray):
            return obj.tolist()
        elif isinstance(obj, TensorDtype):
            return obj.value
        else:
            return super(NpEncoder, self).default(obj)


# TODO: There is only fusion logic, should the logic of splitting tensor be added?
def split_bucket_by_bucket_size(
    tensor_list: List[TensorDeclaration],
    bucket_size: int,
    param_group_info: Dict[str, int] = {},
):
    dtype_unit_size = {
        TensorDtype.F32.value: 4,
        TensorDtype.F16.value: 2,
        TensorDtype.U8.value: 1,
    }

    buckets = []
    tmp_bucket = []
    tmp_bucket_size = 0
    for (dtype, unit_size) in sorted(dtype_unit_size.items()):
        for tensor_declar in [x for x in tensor_list if x["dtype"] == dtype]:
            tensor_size = tensor_declar["num_elements"] * unit_size
            tmp_bucket_size += tensor_size
            tmp_bucket.append(tensor_declar)
            if tmp_bucket_size >= bucket_size:
                buckets.append(tmp_bucket)
                tmp_bucket = []
                tmp_bucket_size = 0

    if len(tmp_bucket) != 0:
        buckets.append(tmp_bucket)

    # group params in bucket
    for i, _ in enumerate(buckets):
        buckets[i] = sorted(
            buckets[i], key=lambda p: param_group_info.get(p["name"], -1)
        )

    return buckets


def record_autotune_log(
    autotune_logfile_path: str,
    autotune_hp: dict,
    train_iter: int,
    score: float
):
    with open(autotune_logfile_path, "a") as autotune_log:
        csv_writer = csv.DictWriter(
            autotune_log,
            fieldnames=sorted(["train_iter", "score"] + list(autotune_hp.keys())),
        )
        first_line = open(autotune_logfile_path).readline()
        if not first_line:
            csv_writer.writeheader()

        cols = copy.deepcopy(autotune_hp)
        cols.update(
            {
                "train_iter": train_iter,
                "score": score,
            }
        )
        cols = OrderedDict(cols)
        logging.info("cols={}".format(cols))
        csv_writer.writerow(cols)


class HyperparameterManager:
    RECORD_MAX_NUM = 1000

    def __init__(
        self,
        is_output_autotune_log: bool,
    ) -> None:
        self.record_deque = collections.deque(
            [
                (
                    -1,
                    BaguaHyperparameter(),
                    float("-inf"),
                )
            ]
        )
        if is_output_autotune_log:
            tmpfile = tempfile.NamedTemporaryFile(
                prefix="bagua_autotune_", mode="w", suffix=".log", delete=False
            )
            self.autotune_logfile_path = tmpfile.name
            tmpfile.close()
        else:
            self.autotune_logfile_path = None

        self.bayesian_optimizer = BayesianOptimizer(
            {
                "bucket_size_2p": IntParam(  # bucket_size = 2 ^ bucket_size_2p
                    val=13,
                    space_dimension=(  # 1KB ~ 2GB
                        10,
                        31,
                    ),
                ),
                "is_hierarchical_reduce": BoolParam(False),
            },
        )

    def tail_record(self) -> Tuple[int, BaguaHyperparameter, float]:
        return self.record_deque[-1]

    def best_hyperparameter(self) -> BaguaHyperparameter:
        return sorted(
            [(score, hp) for (_, hp, score) in self.record_deque],
            key=lambda pair: pair[0],
        )[-1][1]

    def report_metrics(
        self,
        train_iter: int,
        hyperparameter: BaguaHyperparameter,
        system_efficiency_score: float,
    ) -> None:
        while len(self.record_deque) > self.RECORD_MAX_NUM:
            self.record_deque.pop()
        self.record_deque.append(
            (
                train_iter,
                hyperparameter,
                system_efficiency_score,
            )
        )

    def ask_hyperparmeter(
        self,
        train_iter: int,
    ) -> BaguaHyperparameter:
        (_, hp, system_efficiency_score) = self.tail_record()
        optimizer_params = {
            "bucket_size_2p": int(math.log(hp.bucket_size, 2)),
            "is_hierarchical_reduce": hp.is_hierarchical_reduce,
        }
        self.bayesian_optimizer.tell(optimizer_params, system_efficiency_score)
        recommend_param = self.bayesian_optimizer.ask()
        recommend_bucket_size = 2 ** recommend_param["bucket_size_2p"]

        if self.autotune_logfile_path:
            record_autotune_log(
                self.autotune_logfile_path,
                optimizer_params,
                train_iter,
                system_efficiency_score,
            )
        tensor_list = [
            tensor_declar for bucket in hp.buckets for tensor_declar in bucket
        ]
        recommend_buckets = split_bucket_by_bucket_size(
            tensor_list,
            recommend_bucket_size,
        )

        recommend_hp = BaguaHyperparameter(
            buckets=recommend_buckets,
            bucket_size=recommend_bucket_size,
            is_hierarchical_reduce=bool(recommend_param["is_hierarchical_reduce"]),
        )

        return recommend_hp


class AutotuneServiceHyperparameterManager:
    def __init__(self, world_size: int, is_output_autotune_log: bool) -> None:
        self.inner = HyperparameterManager(is_output_autotune_log)
        self.warmup_pass_count = 0
        self.sampling_count = 0
        self.lock = threading.Lock()
        self.check_board = [-1] * world_size
        self.time_hp_last_granted = time.time()
        self.hyperparameter = BaguaHyperparameter()


class AutotuneService:
    def __init__(
        self,
        world_size,
        autotune_level=0,
        max_samples=60,
        sampling_confidence_time_s=5,
        warmup_time_s=30,
        is_output_autotune_log=False,
        default_bucket_size=10 * 1024 ** 2,
    ):
        self.autotune_level = autotune_level
        self.world_size = world_size
        self.max_samples = max_samples
        self.sampling_confidence_time_s = sampling_confidence_time_s
        self.warmup_time_s = warmup_time_s
        self.is_initialized = False
        self.is_output_autotune_log = is_output_autotune_log
        self.default_bucket_size: int = default_bucket_size
        self.model_dict: Dict[str, AutotuneServiceHyperparameterManager] = {}
        self.model_dict_mutex = threading.Lock()

    def autotune(
        self,
        hp_manager: AutotuneServiceHyperparameterManager,
        rank: int,
        train_iter: int,
    ):
        if hp_manager.sampling_count > self.max_samples:
            return

        (
            recommended_train_iter,
            hp,
            system_efficiency_score,
        ) = hp_manager.inner.tail_record()

        logging.info(
            "recommended_train_iter={}, hyperparameters={}, speed={}".format(
                recommended_train_iter,
                hp,
                system_efficiency_score,
            )
        )

        sampling_time = time.time() - hp_manager.time_hp_last_granted

        # Skip at least once during warmup
        if (
            sampling_time < self.warmup_time_s
            or hp_manager.warmup_pass_count == 0  # noqa: W503
        ):
            logging.info(
                "warmup pass, time.time={}, time_hp_last_granted={}, "
                "warmup_time_s={}".format(
                    time.time(),
                    hp_manager.time_hp_last_granted,
                    self.warmup_time_s,
                )
            )
            hp_manager.warmup_pass_count += 1
            return

        if hp_manager.sampling_count == 0:
            sampling_time -= self.warmup_time_s

        if sampling_time < self.sampling_confidence_time_s:
            logging.debug(
                "The sampling time is not up, time={}, last={}, "
                "sampling_confidence_time_s={}".format(
                    time.time(),
                    hp_manager.time_hp_last_granted,
                    self.sampling_confidence_time_s,
                )
            )
            return

        logging.info(
            "rank={}, train_iter={}, sampling_count={}, "
            "max_samples={}".format(
                rank, train_iter, hp_manager.sampling_count, self.max_samples
            )
        )
        recommended_bagua_hp = hp_manager.inner.ask_hyperparmeter(train_iter)
        if hp_manager.sampling_count < self.max_samples:
            hp_manager.hyperparameter = recommended_bagua_hp
        else:
            hp_manager.hyperparameter = hp_manager.inner.best_hyperparameter()

        hp_manager.sampling_count += 1

    def setup_app(self, app):
        @app.route("/api/v1/register_tensors", methods=["POST"])
        def register_tensors():
            req: dict = request.get_json(force=True)
            model_name: str = req["model_name"]
            tensor_list: List[TensorDeclaration] = req["tensor_list"]
            whether_to_bucket: bool = req["whether_to_bucket"]

            with self.model_dict_mutex:
                if model_name not in self.model_dict:
                    self.model_dict[model_name] = AutotuneServiceHyperparameterManager(
                        world_size=self.world_size,
                        is_output_autotune_log=self.is_output_autotune_log,
                    )

            hp_manager = self.model_dict[model_name]
            bucket_size = self.default_bucket_size
            if whether_to_bucket is False:
                bucket_size = 10 * 1024 ** 5

            with hp_manager.lock:
                hp = BaguaHyperparameter(
                    buckets=split_bucket_by_bucket_size(
                        tensor_list,
                        bucket_size,
                    ),
                    bucket_size=bucket_size,
                )
                hp_manager.time_hp_last_granted = time.time()
                hp_manager.hyperparameter = hp
                return json.dumps(
                    {
                        "recommended_hyperparameters": hp.dict(),
                    }
                )

        @app.route("/api/v1/report_metrics", methods=["POST"])
        def report_metrics():
            req: dict = request.get_json(force=True)
            model_name: str = req["model_name"]
            rank: int = req["rank"]
            train_iter: int = req["train_iter"]
            speed: float = req["speed"]
            hyperparameters = req["hyperparameters"]

            if model_name not in self.model_dict:
                return "Service not ready for report_metrics!", 405

            hp_manager = self.model_dict[model_name]

            # Only consider the rank of the first report metrics now.
            with hp_manager.lock:
                (last_report_train_iter, _, _) = hp_manager.inner.tail_record()
                if train_iter <= last_report_train_iter:
                    return json.dumps({})

                logging.debug(
                    "rank={}, train_iter={}, speed={}, "
                    "hyperparameters={}".format(
                        rank,
                        train_iter,
                        speed,
                        hyperparameters,
                    )
                )
                hp_manager.inner.report_metrics(
                    train_iter=train_iter,
                    hyperparameter=BaguaHyperparameter().update(hyperparameters),
                    system_efficiency_score=speed,
                )

            return json.dumps({})

        @app.route("/api/v1/ask_hyperparameters", methods=["POST"])
        def ask_hyperparameters():
            """
            report_metrics must be called before ask_hyperparameters
            """
            req: dict = request.get_json(force=True)
            rank: int = req["rank"]
            model_name: str = req["model_name"]
            train_iter: int = req["train_iter"]

            if model_name not in self.model_dict:
                return "Service not ready for report_metrics!", 405

            hp_manager = self.model_dict[model_name]

            with hp_manager.lock:
                # Autotune conditions:
                # 1. autotune_level >= 1.
                # 2. The bagua process is not in the process of hyperparameter update. (self.check_board.count(self.check_board[0])
                #   == len(self.check_board))
                # 3. Only execute autotune at most once in an iteration. (self.check_board[rank] < train_iter)
                check_board = hp_manager.check_board
                if (
                    self.autotune_level >= 1
                    and check_board.count(check_board[0]) == len(check_board)  # noqa: W503
                    and check_board[rank] < train_iter  # noqa: W503
                ):
                    self.autotune(hp_manager, rank, train_iter)

                check_board[rank] = train_iter

                return json.dumps(
                    {
                        "recommended_hyperparameters": hp_manager.hyperparameter.dict(),
                        "is_autotune_completed": hp_manager.sampling_count
                        > self.max_samples,  # noqa: W503
                    }
                )

        # set secret-key
        app.config.update(SECRET_KEY=os.urandom(24))

        return app


class AutotuneClient:
    def __init__(
        self,
        service_addr: str,
        service_port: int,
        proxies={
            "http": None,
            "https": None,
        },
    ):
        self.autotune_service_addr = "{}:{}".format(service_addr, service_port)
        self.session = requests.Session()
        self.proxies = proxies

    def report_metrics(
        self,
        model_name: str,
        rank: int,
        train_iter: int,
        hyperparameters: dict,
        speed: float,
    ) -> requests.Response:
        rsp = self.session.post(
            "http://{}/api/v1/report_metrics".format(self.autotune_service_addr),
            json={
                "model_name": model_name,
                "rank": rank,
                "train_iter": train_iter,
                "hyperparameters": hyperparameters,
                "speed": speed,
            },
            proxies=self.proxies,
        )
        return rsp

    def register_tensors(
        self,
        model_name: str,
        tensor_list: List[TensorDeclaration],
        whether_to_bucket: bool = True,
    ) -> requests.Response:
        rsp = self.session.post(
            "http://{}/api/v1/register_tensors".format(self.autotune_service_addr),
            json={
                "model_name": model_name,
                "tensor_list": tensor_list,
                "whether_to_bucket": whether_to_bucket,
            },
            proxies=self.proxies,
        )
        return rsp

    def ask_hyperparameters(
        self,
        model_name: str,
        rank: int,
        train_iter: int,
    ) -> requests.Response:
        rsp = self.session.post(
            "http://{}/api/v1/ask_hyperparameters".format(self.autotune_service_addr),
            json={
                "model_name": model_name,
                "rank": rank,
                "train_iter": train_iter,
            },
            proxies=self.proxies,
        )
        return rsp


if __name__ == "__main__":
    autotune_service = AutotuneService(10)
    app = Flask(__name__)
    app = autotune_service.setup_app(app)

    server = multiprocessing.Process(
        target=app.run,
        kwargs={
            "host": "0.0.0.0",
            "port": 8123,
        },
    )
    server.daemon = True
    server.start()
    server.join()



import unittest
import logging
import multiprocessing
from flask import Flask
from bagua.bagua_define import TensorDeclaration
from bagua.service import AutotuneService, AutotuneClient
from bagua.bagua_define import BaguaHyperparameter, get_tensor_declaration_bytes
import socket
import numpy as np


def pick_n_free_ports(n: int):
    socks = []
    for i in range(n):
        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        sock.bind(("localhost", 0))
        socks.append(sock)

    n_free_ports = [sock.getsockname()[1] for sock in socks]
    for sock in socks:
        sock.close()

    return n_free_ports


def metrics(buckets, is_hierarchical_reduce):
    # score = sum(-abs(bucket_sum_size - 20MB))
    # convex function with peak at bucket_size=20MB
    score = 0.0
    for bucket in buckets:
        score += -abs(sum([get_tensor_declaration_bytes(td)
                      for td in bucket]) - 20 * 1024 ** 2)

    if not is_hierarchical_reduce:
        score += abs(score) * 0.1

    return score


class MockBaguaProcess:
    def __init__(
        self,
        rank,
        service_addr,
        service_port,
        model_name,
        tensor_list,
    ) -> None:
        self.rank = rank
        self.model_name = model_name
        self.tensor_list = tensor_list
        self.client = AutotuneClient(service_addr, service_port)

    def run(self, total_iters=5000):
        rsp = self.client.register_tensors(self.model_name, self.tensor_list)
        assert rsp.status_code == 200, "register_tensors failed, rsp={}".format(
            rsp)
        hp = BaguaHyperparameter().update(
            rsp.json()["recommended_hyperparameters"])

        for train_iter in range(total_iters):
            score = metrics(hp.buckets, hp.is_hierarchical_reduce)
            rsp = self.client.report_metrics(
                self.model_name, self.rank, train_iter, hp.dict(), score
            )
            assert rsp.status_code == 200, "report_metrics failed, rsp={}".format(
                rsp)
            rsp = self.client.ask_hyperparameters(
                self.model_name, self.rank, train_iter
            )
            assert rsp.status_code == 200, "ask_hyperparameters failed, rsp={}".format(
                rsp
            )
            rsp = rsp.json()
            hp.update(rsp["recommended_hyperparameters"])

            if rsp["is_autotune_completed"]:
                logging.info("train_iter={}".format(train_iter))
                break

        return hp


class TestAutotuneService(unittest.TestCase):
    def test_autotune_service(self):
        service_addr = "127.0.0.1"
        service_port = pick_n_free_ports(1)[0]
        nprocs = 2
        np.random.seed(123)

        autotune_service = AutotuneService(
            nprocs,
            autotune_level=1,
            sampling_confidence_time_s=0.1,
            warmup_time_s=1.0,
            is_output_autotune_log=True,
        )
        app = Flask(__name__)
        app = autotune_service.setup_app(app)
        log = logging.getLogger("werkzeug")
        log.setLevel(logging.ERROR)

        server = multiprocessing.Process(
            target=app.run,
            kwargs={
                "host": "0.0.0.0",
                "port": service_port,
                "debug": False,
            },
        )
        server.daemon = True
        server.start()

        model_dict = {
            "m1": [
                TensorDeclaration(
                    {
                        "name": "A",
                        "num_elements": 1 * 1024 ** 2,
                        "dtype": "f32",
                    }
                ),
                TensorDeclaration(
                    {
                        "name": "B",
                        "num_elements": 2 * 1024 ** 2,
                        "dtype": "f32",
                    }
                ),
                TensorDeclaration(
                    {
                        "name": "C",
                        "num_elements": 3 * 1024 ** 2,
                        "dtype": "f32",
                    }
                ),
                TensorDeclaration(
                    {
                        "name": "D",
                        "num_elements": 4 * 1024 ** 2,
                        "dtype": "f32",
                    }
                ),
                TensorDeclaration(
                    {
                        "name": "E",
                        "num_elements": 5 * 1024 ** 2,
                        "dtype": "f32",
                    }
                ),
            ],
            "m2": [
                TensorDeclaration(
                    {
                        "name": "A",
                        "num_elements": 1 * 1024 ** 2,
                        "dtype": "f32",
                    }
                ),
                TensorDeclaration(
                    {
                        "name": "B",
                        "num_elements": 3 * 1024 ** 2,
                        "dtype": "f32",
                    }
                ),
                TensorDeclaration(
                    {
                        "name": "C",
                        "num_elements": 5 * 1024 ** 2,
                        "dtype": "f16",
                    }
                ),
                TensorDeclaration(
                    {
                        "name": "D",
                        "num_elements": 7 * 1024 ** 2,
                        "dtype": "f16",
                    }
                ),
                TensorDeclaration(
                    {
                        "name": "E",
                        "num_elements": 11 * 1024 ** 2,
                        "dtype": "f32",
                    }
                ),
            ],
        }

        mock_objs = []
        with multiprocessing.pool.ThreadPool(nprocs * len(model_dict)) as pool:
            results = dict([(key, []) for key in model_dict.keys()])
            for i in range(nprocs):
                for (model_name, tensor_list) in model_dict.items():
                    mock = MockBaguaProcess(
                        i, service_addr, service_port, model_name, tensor_list
                    )
                    mock_objs.append(mock)
                    ret = pool.apply_async(mock.run)
                    results[model_name].append(ret)
            for ret in results["m1"]:
                hp = ret.get()
                buckets = [[
                    td["name"] for td in bucket] for bucket in hp.buckets]
                self.assertEqual(
                    buckets,
                    [['A', 'B', 'C'], ['D'], ['E']],
                    "hp={}".format(hp.dict())
                )
            for ret in results["m2"]:
                hp = ret.get()
                buckets = [[
                    td["name"] for td in bucket] for bucket in hp.buckets]
                self.assertEqual(
                    buckets,
                    [['C', 'D'], ['A', 'B'], ['E']],
                    "hp={}".format(hp.dict())
                )

        server.terminate()
        server.join()


if __name__ == "__main__":
    logging.basicConfig(
        level=logging.WARN,
        format="%(filename)s: %(levelname)s: %(funcName)s(): %(lineno)d:\t%(message)s",
    )
    unittest.main()



import torch
from torch.utils.data import TensorDataset
from bagua.torch_api.tools.data import (
    LoadBalancingDistributedSampler,
    LoadBalancingDistributedBatchSampler,
)
import logging


logging.basicConfig(level=logging.DEBUG)


def test_loadbalance():
    n = 100
    dataset = TensorDataset(torch.randn(n, 2), torch.arange(n))
    sampler = LoadBalancingDistributedSampler(
        dataset, lambda x: 2 ** (x[1].item() - 1).bit_length(), num_replicas=2, rank=0
    )

    dataloader = torch.utils.data.DataLoader(
        dataset,
        batch_size=4,
        shuffle=False,
        sampler=sampler,
    )

    for i, data in enumerate(dataloader):
        print(i, data)


def test_batchsampler():
    n = 20
    dataset = TensorDataset(torch.randn(n, 2), torch.arange(n))

    import random

    def batch_fn(indices):
        batch = []
        batch_size = random.randint(1, 4)
        for i in indices:
            batch.append(i)
            if len(batch) == batch_size:
                yield batch
                batch = []
                batch_size = random.randint(1, 4)
        if len(batch) > 0:
            yield batch

    sampler = LoadBalancingDistributedSampler(
        dataset,
        lambda x: 2 ** (x[1].item() - 1).bit_length(),
        num_replicas=2,
        rank=0,
    )

    batch_sampler = LoadBalancingDistributedBatchSampler(
        sampler,
        batch_fn=lambda x: list(batch_fn(x)),
    )

    dataloader = torch.utils.data.DataLoader(dataset, batch_sampler=batch_sampler)

    for epoch in range(10):
        print("Training ....")
        batch_sampler.set_epoch(epoch)

        for i, data in enumerate(dataloader):
            print(i, data)


if __name__ == "__main__":
    test_loadbalance()
    test_batchsampler()



#!/usr/bin/env python3
"""
Utilities for QSGD algorithm.
"""
from torch.distributions.bernoulli import Bernoulli
import torch
import cupy
from bagua.torch_api.utils import torch2cupy, cupy2torch


def qsgd_compress(t, pnorm="inf", quan_bits=8):
    """
    Quantize float32 into uint8 using QSGD algorithm.

    Arguments:
        * `t` - Input PyTorch CUDA Tensor to be quantized.
        * `pnorm` - Order of norm used for QSGD compression. Default value is `inf`.
        * `quan_bits` - Number of quantization bits used for QSGD compression. Default
           value is `8`. Other values like `1`,`2`,`4` are not supported at present.

    Yields:
        (Torch.tensor, Torch.tensor, Torch.tensor):  Quantized PyTorch CUDA Tensors.

    Examples:

    ```python
    cupy_stream = cupy.cuda.ExternalStream(torch.cuda.current_stream().cuda_stream)
    cupy_stream.use()
    norm, sign, compressed_ints = qsgd_compress(tensor)
    ```

    ..note::
        CuPy and PyTorch use different default CUDA streams. Force CuPy to use PyTorch
        current CUDA stream to simplify stream synchronization.

    """
    _quantization_level = (1 << quan_bits) - 1

    norm = t.norm(float(pnorm))

    sign = (torch.sign(t) + 1.0).bool()

    _level = t / norm * _quantization_level
    _bernoulli_probs = _level - _level.floor()
    _incr = Bernoulli(probs=_bernoulli_probs).sample()
    _compressed_floats = (_level.floor() + _incr.float()).abs()
    compressed_floats = torch.clamp(_compressed_floats, max=_quantization_level)
    compressed_ints = compressed_floats.byte()

    packed_sign = _cupy_packbits(sign)
    return norm, packed_sign, compressed_ints


def qsgd_decompress(norm, packed_sign, compressed_ints, quan_bits=8):
    """
    The reverse of the ``qsgd_compress`` function.

    Arguments:
        * `norm` - Order of norm used in QSGD compression.
        * `packed_sign` - Packed sign of quantized tensor.
        * `compressed_ints` - Absolute value of quantized tensor.
        * `quan_bits` - Number of quantization bits. Default value is 8. Other
           values like `1`,`2`,`4` is not supported at present.

    Yields:
        Torch.tensor: De-quantized tensor.

    Examples:

    ```python
    cupy_stream = cupy.cuda.ExternalStream(torch.cuda.current_stream().cuda_stream)
    cupy_stream.use()
    tensor = qsgd_compress(norm, sign, compressed_ints)
    ```

    ..note::
        CuPy and PyTorch use different default CUDA streams. Force CuPy to use PyTorch
        current CUDA stream to simplify stream synchronization.
    """

    compressed_floats = compressed_ints.float()
    _quantization_level = (1 << quan_bits) - 1
    sign = _cupy_unpackbits(packed_sign, compressed_floats.size(0))
    return norm * (sign.float() * 2 - 1) * compressed_floats / _quantization_level


def _cupy_packbits(tensor):
    cupy_tensor = torch2cupy(tensor)
    packed_cupy_tensor = cupy.packbits(cupy_tensor)
    packed_tensor = cupy2torch(packed_cupy_tensor)
    return packed_tensor


def _cupy_unpackbits(tensor, size):
    cupy_tensor = torch2cupy(tensor)
    unpacked_cupy_tensor = cupy.unpackbits(cupy_tensor)[0:size]
    tensor = cupy2torch(unpacked_cupy_tensor)
    return tensor



from bagua.torch_api.distributed import BaguaModule
from bagua.torch_api.bucket import BaguaBucket
from bagua.torch_api.tensor import BaguaTensor
from typing import List
import torch


class Algorithm:
    """
    This is the base class that all Bagua algorithms inherit.

    It provides methods that can be override to implement different kinds of
    distributed algorithms.
    """

    def need_reset(self) -> bool:
        """
        Returns:
            True if all initialization methods of the current algorithms should
            be called again. This is useful for algorithms that has multiple
            stages where each stage needs different initializations.
        """
        return False

    def init_tensors(self, bagua_module: BaguaModule) -> List[BaguaTensor]:
        """
        Given a BaguaModule, return Bagua tensors to be used in Bagua for later
        operations.

        Args:
            bagua_module: A PyTorch module initialized by
                ``with_bagua(...)`` method.

        Returns:
            A list of Bagua tensors for communication.
        """

        parameters = bagua_module.bagua_build_params()
        tensors = []
        for name, param in parameters.__reversed__():
            grad = param.bagua_ensure_grad().ensure_bagua_tensor(
                name, bagua_module.bagua_module_name
            )
            param._bagua_grad = grad
            tensors.append(grad)
        self._communication_tensor_names = set(name for name, _ in parameters)
        assert len(self._communication_tensor_names) == len(
            tensors
        ), "tensor names should be unique"
        return tensors

    def tensors_to_buckets(self, tensors: List[List[BaguaTensor]]) -> List[BaguaBucket]:
        """
        Given the bucketing suggestion from Bagua, return the actual Bagua buckets.
        The default implementation follows the suggestion to do the bucketing.

        Args:
            tensors: Bagua tensors grouped in different
                lists, representing Bagua's suggestion on how to bucketing the
                tensors.

        Returns:
            A list of Bagua buckets.
        """
        bagua_buckets = []
        for idx, bucket in enumerate(tensors):
            bagua_bucket = BaguaBucket(
                bucket, flatten=True, name=str(idx)
            )  # TODO: check duplicated names
            bagua_buckets.append(bagua_bucket)
        return bagua_buckets

    def init_forward_pre_hook(self, bagua_module: BaguaModule):
        """Given a `BaguaModule`, return a hook function that will be executed before the
        forward process.

        Args:
            bagua_module: A PyTorch module initialized by
                ``with_bagua(...)`` method.

        Returns:
            A function that takes the model's input.
        """

        def hook(input):
            pass

        return hook

    def init_backward_hook(self, bagua_module: BaguaModule):
        """Given a `BaguaModule`, return a hook function that will be executed on every
        parameter's gradient computation completion.

        Args:
            bagua_module: A PyTorch module initialized by
                ``with_bagua(...)`` method.

        Returns:
            A function that takes the name of a parameter (as in
            torch.nn.Module.named_parameters()) and the parameter itself.
        """

        def hook(parameter_name, parameter):
            if parameter_name in self._communication_tensor_names:
                assert (
                    parameter._bagua_grad.data_ptr() == parameter.grad.data_ptr()
                ), "bagua grad data_ptr should match parameter grad"
                parameter._bagua_grad.bagua_mark_communication_ready()

        return hook

    def init_post_backward_hook(self, bagua_module: BaguaModule):
        """Given a `BaguaModule`, return a hook function that will be executed when the
        backward pass is done.

        Args:
            bagua_module: A PyTorch module initialized by
                ``with_bagua(...)`` method.

        Returns:
            A function that takes no argument.
        """

        def hook():
            bagua_module._bagua_backend.wait_pending_comm_ops()

        return hook

    def init_post_optimizer_step_hook(self, bagua_module: BaguaModule):
        """Given a `BaguaModule`, return a hook function that will be executed when the
        ``optimizer.step()`` is done.

        Args:
            bagua_module: A PyTorch module initialized by
                ``with_bagua(...)`` method.

        Returns:
            A function that takes the optimizer that is called step().
        """

        def hook(optimizer: torch.optim.Optimizer):
            pass

        return hook

    def init_operations(
        self,
        bagua_module: BaguaModule,
        bucket: BaguaBucket,
    ):
        """Given a `BaguaModule`, and a Bagua bucket, register operations to be
        executed on the bucket.

        Args:
            bagua_module: A PyTorch module initialized by
                ``with_bagua(...)`` method.
            bucket: A single bucket to register operations.
        """
